/* Copyright (C) 1989 Aladdin Enterprises.  All rights reserved.
   Distributed by Free Software Foundation, Inc.

This file is part of Ghostscript.

Ghostscript is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves any
particular purpose or works at all, unless he says so in writing.  Refer
to the Ghostscript General Public License for full details.

Everyone is granted permission to copy, modify and redistribute
Ghostscript, but only under the conditions described in the Ghostscript
General Public License.  A copy of this license is supposed to have been
given to you along with Ghostscript so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies.  */

/* gsdevice.c */
/* Device operators for GhostScript library */
#include <math.h>			/* for fabs */
#include "gx.h"
#include "gserrors.h"
#include "gxfixed.h"			/* ditto */
#include "gxmatrix.h"			/* for gzstate.h */
#include "gzstate.h"
#include "gzdevice.h"

/* Device definitions */
/* Following defines the null device */
int gx_default_open_device(P1(gx_device *));
int gx_default_close_device(P1(gx_device *));
int gx_default_sync_output(P1(gx_device *));
int gx_default_map_rgb_color(P4(gx_device *, ushort, ushort, ushort));
int gx_default_map_color_rgb(P3(gx_device *, int, ushort *));
int null_fill_rectangle(P6(gx_device *, int, int, int, int, int));
int null_tile_rectangle(P8(gx_device *, gx_bitmap *, int, int, int, int, int, int));
int null_copy_mono(P10(gx_device *, byte *, int, int, int, int, int, int, int, int));
int null_copy_color(P9(gx_device *, byte *, int, int, int, int, int, int, int));
int null_draw_line(P6(gx_device *, int, int, int, int, int));
int null_fill_trapezoid(P8(gx_device *, int, int, int, int, int, int, int));
int null_tile_trapezoid(P10(gx_device *, gx_bitmap *, int, int, int, int, int, int, int, int));
static gx_device_procs null_procs = {
	gx_default_open_device,
	gx_default_close_device,
	gx_default_map_rgb_color,
	gx_default_map_color_rgb,
	gx_default_sync_output,
	null_fill_rectangle,
	null_tile_rectangle,
	null_copy_mono,
	null_copy_color,
	null_draw_line,
	null_fill_trapezoid,
	null_tile_trapezoid
};
static gx_device null_device = {
	sizeof(device),
	&null_procs,
	0x3fff, 0x3fff,			/* 0, 0, */
	0, 1, 1, 1
};
/* The default (only) frame device. */
/* External initialization code must set this up. */
gx_device *gx_device_default_p = &null_device;

/* Synchronize the device with the accumulated page description */
int
gs_copypage(gs_state *pgs)
{	gx_device *dev = pgs->device->info;
	return (*dev->procs->sync_output)(dev);
}

/* Set up the only recognized device */
int
gs_framedevice(gs_state *pgs, gs_matrix *pmat, int width, int height)
{	device *pdev = pgs->device;
	gx_device *dev = gx_device_default_p;
	if ( width <= 0 || height <= 0 ) return_error(gs_error_rangecheck);
	pdev->info = dev;
	pdev->initial_transformation = *pmat;
	pdev->upright = ((pmat->xx < 0) == (pmat->yy < 0));
	pdev->aspect_ratio = fabs(pmat->xx / pmat->yy);
	/* Initialize the device */
	(*dev->procs->open_device)(dev);
	/* Compute device white and black codes */
	pdev->black = (*dev->procs->map_rgb_color)(dev, 0, 0, 0);
	pdev->white = (*dev->procs->map_rgb_color)(dev, dev->max_rgb_value, dev->max_rgb_value, dev->max_rgb_value);
	return 0;
}

/* Select the null device */
void
gs_nulldevice(gs_state *pgs)
{	void gx_device_no_output(P1(gs_state *));
	device *pdev = pgs->device;
	pdev->info = &null_device;
	gs_make_identity(&pdev->initial_transformation);
	pdev->is_band_device = 0;
}

/* Read out the current device */
void
gs_currentdevice(gs_state *pgs, gs_matrix *pmat, int *pwidth, int *pheight)
{	device *pdev = pgs->device;
	gx_device *info = pdev->info;
	*pmat = pdev->initial_transformation;
	*pwidth = info->width / 8;
	*pheight = info->height;
}

/* Install enough of a null device to suppress graphics output */
/* during the execution of stringwidth. */
void
gx_device_no_output(gs_state *pgs)
{	pgs->device->info = &null_device;
}

/* Dummy device procedures */
static int
null_fill_rectangle(gx_device *dev, int x, int y, int w, int h, int color)
{	return 0;
}
static int
null_tile_rectangle(gx_device *dev, gx_bitmap *tile,
  int x, int y, int w, int h, int zero, int one)
{	return 0;
}
static int
null_copy_mono(gx_device *dev, byte *data,
  int dx, int raster, int x, int y, int w, int h, int zero, int one)
{	return 0;
}
static int
null_copy_color(gx_device *dev, byte *data,
  int dx, int raster, int x, int y, int w, int h, int transparent)
{	return 0;
}
static int
null_draw_line(gx_device *dev, int x0, int y0, int x1, int y1, int color)
{	return 0;
}
static int
null_fill_trapezoid(gx_device *dev,
  int x0, int y0, int w0, int x1, int y1, int w1, int color)
{	return 0;
}
static int
null_tile_trapezoid(gx_device *dev, gx_bitmap *tile,
  int x0, int y0, int w0, int x1, int y1, int w1, int color0, int color1)
{	return 0;
}
/* Default (no-op) device open, close, map, and sync procedures */
int
gx_default_open_device(gx_device *dev)
{	return 0;
}
int
gx_default_close_device(gx_device *dev)
{	return 0;
}
int
gx_default_map_rgb_color(gx_device *dev, ushort r, ushort g, ushort b)
{	return max(max(r, g), b);
}
int
gx_default_map_color_rgb(gx_device *dev, int color, ushort *prgb)
{	prgb[0] = prgb[1] = prgb[2] = color;
	return 0;
}
int
gx_default_sync_output(gx_device *dev)
{	return 0;
}

#ifdef gs_DEBUG

/* ------ Tracing 'device' ------*/

/* Device structure */
typedef struct gx_device_trace_s {
	gx_device_common;
	gx_device *real_device;		/* real device */
} gx_device_trace;
#define trdev (((gx_device_trace *)dev)->real_device)

/* Procedure structure */
int trace_open_device(P1(gx_device *));
int trace_close_device(P1(gx_device *));
int trace_sync_output(P1(gx_device *));
int trace_map_rgb_color(P4(gx_device *, ushort, ushort, ushort));
int trace_map_color_rgb(P3(gx_device *, int, ushort *));
int trace_fill_rectangle(P6(gx_device *, int, int, int, int, int));
int trace_tile_rectangle(P8(gx_device *, gx_bitmap *, int, int, int, int, int, int));
int trace_copy_mono(P10(gx_device *, byte *, int, int, int, int, int, int, int, int));
int trace_copy_color(P9(gx_device *, byte *, int, int, int, int, int, int, int));
int trace_draw_line(P6(gx_device *, int, int, int, int, int));
int trace_fill_trapezoid(P8(gx_device *, int, int, int, int, int, int, int));
int trace_tile_trapezoid(P10(gx_device *, gx_bitmap *, int, int, int, int, int, int, int, int));
static gx_device_procs trace_procs = {
	trace_open_device,
	trace_close_device,
	trace_map_rgb_color,
	trace_map_color_rgb,
	trace_sync_output,
	trace_fill_rectangle,
	trace_tile_rectangle,
	trace_copy_mono,
	trace_copy_color,
	trace_draw_line,
	trace_fill_trapezoid,
	trace_tile_trapezoid
};

/* Make a traced version of a device. */
gx_device *
gs_trace_device(gx_device *rdev)
{	gx_device_trace *dev = (gx_device_trace *)gs_malloc(sizeof(gx_device_trace));
	if ( dev == 0 ) return 0;
	*(gx_device *)dev = *rdev;	/* copy common info */
	dev->params_size = sizeof(gx_device_trace);
	dev->procs = &trace_procs;
	dev->real_device = rdev;
	return (gx_device *)dev;
}

/* Utilities */
static int
print_code(int result)
{
if ( gs_debug['v'] )
  {	if ( result == 0 )
		printf(";\n");
	else
		printf(";	/* = %d */\n", result);
  }
	return result;
}

/* Procedures */
static int
trace_open_device(gx_device *dev)
{	gx_device *rdev = trdev;
	int result = (*rdev->procs->open_device)(rdev);
if ( gs_debug['v'] )
	printf("\tT(open_device)(dev)");
	return print_code(result);
}
static int
trace_close_device(gx_device *dev)
{	gx_device *rdev = trdev;
	int result = (*rdev->procs->close_device)(rdev);
if ( gs_debug['v'] )
	printf("\tT(close_device)(dev)");
	return print_code(result);
}
static int
trace_map_rgb_color(gx_device *dev, ushort r, ushort g, ushort b)
{	gx_device *rdev = trdev;
	int result = (*rdev->procs->map_rgb_color)(rdev, r, g, b);
if ( gs_debug['v'] )
	printf("\tT(map_rgb_color)(dev, %u, %u, %u) /*= %d */;\n", r, g, b, result);
	return result;
}
static int
trace_map_color_rgb(gx_device *dev, int color, ushort *prgb)
{	gx_device *rdev = trdev;
	int result = (*rdev->procs->map_color_rgb)(rdev, color, prgb);
if ( gs_debug['v'] )
	printf("\t{ ushort rgb[3]; T(map_color_rgb)(dev, %d, rgb /* %u, %u, %u */); }",
		color, prgb[0], prgb[1], prgb[2]);
	return print_code(result);
}
static int
trace_sync_output(gx_device *dev)
{	gx_device *rdev = trdev;
	int result = (*rdev->procs->sync_output)(rdev);
if ( gs_debug['v'] )
	printf("\tT(sync_output)(dev)");
	return print_code(result);
}
static int
trace_fill_rectangle(gx_device *dev, int x, int y, int w, int h, int color)
{	gx_device *rdev = trdev;
	int result = (*rdev->procs->fill_rectangle)(rdev, x, y, w, h, color);
if ( gs_debug['v'] )
	printf("\tT(fill_rectangle)(dev, %d, %d, %d, %d, %d)", x, y, w, h, color);
	return print_code(result);
}
static int
trace_tile_rectangle(gx_device *dev, gx_bitmap *tile,
  int x, int y, int w, int h, int zero, int one)
{	gx_device *rdev = trdev;
	int result = (*rdev->procs->tile_rectangle)(rdev, tile,
		x, y, w, h, zero, one);
if ( gs_debug['v'] )
   {	int i;
	printf("\t{ static byte data = { 0x%x", tile->data[0]);
	for ( i = 1; i < tile->raster * tile->height; i++ )
		printf(", 0x%x", tile->data[i]);
	printf(" };\n\t  static gx_bitmap tile = { &data, %d, %d, %d };\n",
		tile->raster, tile->width, tile->height);
	printf("\t  T(tile_rectangle)(dev, &tile, %d, %d, %d, %d, %d, %d);\n\t}",
		x, y, w, h, zero, one);
   }
	return print_code(result);
}
static int
trace_copy_mono(gx_device *dev, byte *data,
  int dx, int raster, int x, int y, int w, int h, int zero, int one)
{	gx_device *rdev = trdev;
	int result = (*rdev->procs->copy_mono)(rdev, data,
		dx, raster, x, y, w, h, zero, one);
if ( gs_debug['v'] )
	printf("\tT(copy_mono)(dev, data, %d, %d, %d, %d, %d, %d, %d, %d)",
		dx, raster, x, y, w, h, zero, one);
	return print_code(result);
}
static int
trace_copy_color(gx_device *dev, byte *data,
  int dx, int raster, int x, int y, int w, int h, int transparent)
{	gx_device *rdev = trdev;
	int result = (*rdev->procs->copy_color)(rdev, data,
		dx, raster, x, y, w, h, transparent);
if ( gs_debug['v'] )
	printf("\tT(copy_color)(dev, data, %d, %d, %d, %d, %d, %d, %d)",
		dx, raster, x, y, w, h, transparent);
	return print_code(result);
}
static int
trace_draw_line(gx_device *dev, int x0, int y0, int x1, int y1, int color)
{	gx_device *rdev = trdev;
	int result = (*rdev->procs->draw_line)(rdev, x0, y0, x1, y1, color);
if ( gs_debug['v'] )
	printf("\tT(draw_line)(dev, %d, %d, %d, %d, %d)",
		x0, y0, x1, y1, color);
	return print_code(result);
}
static int
trace_fill_trapezoid(gx_device *dev,
  int x0, int y0, int w0, int x1, int y1, int w1, int color)
{	gx_device *rdev = trdev;
	int result = (*rdev->procs->fill_trapezoid)(rdev,
		x0, y0, w0, x1, y1, w1, color);
if ( gs_debug['v'] )
	printf("\tT(fill_trapezoid)(dev, %d, %d, %d, %d, %d, %d, %d)",
		x0, y0, w0, x1, y1, w1, color);
	return print_code(result);
}
static int
trace_tile_trapezoid(gx_device *dev, gx_bitmap *tile,
  int x0, int y0, int w0, int x1, int y1, int w1, int color0, int color1)
{	gx_device *rdev = trdev;
	int result = (*rdev->procs->tile_trapezoid)(rdev, tile,
		x0, y0, w0, x1, y1, w1, color0, color1);
if ( gs_debug['v'] )
	printf("\tT(tile_trapezoid)(dev, %d, %d, %d, %d, %d, %d, %d, %d)",
		x0, y0, w0, x1, y1, w1, color0, color1);
	return print_code(result);
}

#endif
